theory Complexity 
imports Show Term
begin

definition deriv_bound :: "'a ars \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> bool"
  where "deriv_bound r a n \<equiv> \<not> (\<exists> b. (a,b) \<in> (r ^^ Suc n))"


lemma deriv_bound_empty[simp]: "deriv_bound {} a n"
  unfolding deriv_bound_def by auto

lemma deriv_bound_image: 
  assumes b: "deriv_bound r' (f a) n"
  and step: "\<And> a b. (a,b) \<in> r \<Longrightarrow> (f a, f b) \<in> r'"
  shows "deriv_bound r a n"
proof -
  note d = deriv_bound_def
  show ?thesis unfolding d
  proof
    assume "\<exists> b. (a,b) \<in> r ^^ Suc n"
    then obtain b where "(a,b) \<in> r ^^ Suc n" ..
    from this[unfolded rel_pow_fun_conv]
    obtain g where g0: "g 0 = a" and last: "g (Suc n) = b"
      and steps: "\<And> i. i < Suc n \<Longrightarrow> (g i, g (Suc i)) \<in> r" by auto
    {
      fix i
      assume "i < Suc n"
      from step[OF steps[OF this]] have "(f (g i), f (g (Suc i))) \<in> r'" .
    } note steps = this
    with g0 last have "(f a, f b) \<in> r' ^^ Suc n"
      unfolding rel_pow_fun_conv
      by (intro exI[of _ "\<lambda> i. f (g i)"], auto)
    with b[unfolded d] show False by auto
  qed
qed

lemma deriv_bound_subset: assumes subset: "r \<subseteq> r'"
  and b: "deriv_bound r' a n"
  shows "deriv_bound r a n"
  by (rule deriv_bound_image[of _ "\<lambda> x. x", OF b], insert subset, auto)

lemma deriv_bound_mono: assumes "deriv_bound r a n"
  and "m \<ge> n"
  shows "deriv_bound r a m"
proof -
  note d = deriv_bound_def
  from assms(2) have "m = n + (m - n)" by auto
  then obtain k where m: "m = n + k" by blast
  show ?thesis unfolding d
  proof
    assume "\<exists> b. (a,b) \<in> r ^^ Suc m" 
    then obtain b where "(a,b) \<in> r ^^ (Suc n + k)" unfolding m by auto
    from this[unfolded rel_pow_add] obtain b where "(a,b) \<in> r ^^ Suc n" by auto
    with assms(1)[unfolded d] show False by blast
  qed
qed

lemma deriv_bound_SN_on: assumes b: "deriv_bound r a n"
  shows "SN_on r {a}"
proof
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> r"
  and f0: "f 0 \<in> {a}"
  with b[unfolded deriv_bound_def]
  have not: "(f 0, f (Suc n)) \<notin> r ^^ Suc n" by blast
  have "(f 0, f (Suc n)) \<in> r ^^ Suc n" unfolding rel_pow_fun_conv
    by (rule exI[of _ f], insert steps, auto)
  with not show False ..
qed

lemma deriv_bound_steps: assumes steps: "(a,b) \<in> r^^n"
  and bound: "deriv_bound r a m"
  shows "n \<le> m"
proof (rule ccontr)
  assume "\<not> ?thesis"
  hence n: "n \<ge> Suc m" by auto
  from steps[unfolded rel_pow_fun_conv] obtain f where 
    f0: "f 0 = a"  and steps: "\<And> i. i < n \<Longrightarrow> (f i, f (Suc i)) \<in> r" by auto
  {
    fix i
    assume "i < Suc m"
    with n have "i < n" by simp
    note steps[OF this]
  }
  with f0 have "(a,f (Suc m)) \<in> r ^^ (Suc m)" unfolding rel_pow_fun_conv by blast
  with bound[unfolded deriv_bound_def] show False by blast
qed

datatype complexity_class = Comp_Poly nat

fun complexity_of :: "complexity_class \<Rightarrow> (nat \<Rightarrow> nat)set"
  where  "complexity_of (Comp_Poly d) = { \<lambda> n. c * n ^ d + e | c e. True}"

lemma complexity_poly_mono_deg: assumes d: "d \<le> d'" 
  and f: "f \<in> complexity_of (Comp_Poly d)"
  shows "\<exists> g. g \<in> complexity_of (Comp_Poly d') \<and> (\<forall> n. f n \<le> g n)"
proof -
  from f obtain c e where f: "f = (\<lambda> n. c * n ^ d + e)" by auto
  let ?g = "\<lambda> n. c * n ^ d' + e"
  {
    fix n
    have "f n \<le> ?g n + c"
    proof (cases n)
      case (Suc m)
      have "f n \<le> ?g n" 
        using power_increasing[OF d, of n] 
        unfolding f Suc by auto
      thus ?thesis by auto
    next
      case 0
      show ?thesis unfolding f 0
        by (cases d, auto)
    qed
  } note main = this
  obtain g where g: "g = (\<lambda>n. ?g n + c)" by auto
  from main have le: "\<forall> n. f n \<le> g n" unfolding g by simp
  have g: "g \<in> complexity_of (Comp_Poly d')"
    unfolding g complexity_of.simps
    by (rule, rule exI[of _ c], rule exI[of _ "e + c"], auto)
  show ?thesis using le g by auto
qed

lemma complexity_poly_mono: 
  assumes f: "f \<in> complexity_of (Comp_Poly d)"
  and nm: "n \<le> m"
  shows "f n \<le> f m"
proof -
  from f obtain c e where f: "f = (\<lambda> n. c * n ^ d + e)" by auto
  show ?thesis unfolding f
    using power_mono[OF nm, of d]
    by auto
qed

definition O_of :: "complexity_class \<Rightarrow> (nat \<Rightarrow> nat)set"
where "O_of cl \<equiv> {f. \<exists> g \<in> complexity_of cl. (\<forall> n. f n \<le> g n)}"

lemma O_of_poly_mono_deg: assumes d: "d \<le> d'" 
  shows "O_of (Comp_Poly d) \<subseteq> O_of (Comp_Poly d')"
proof 
  fix f
  assume "f \<in> O_of (Comp_Poly d)"
  then obtain g where g: "g \<in> complexity_of (Comp_Poly d)" and fg: "\<And> n. f n \<le> g n" 
    unfolding O_of_def by auto
  from complexity_poly_mono_deg[OF d g]
  obtain h where h: "h \<in> complexity_of (Comp_Poly d')" and gh: "\<And> n. g n \<le> h n" 
    by auto
  from le_trans[OF fg gh] h
  show "f \<in> O_of (Comp_Poly d')" unfolding O_of_def by blast
qed

lemma O_of_polyI[intro]: 
  assumes ge: "\<And> n. f n \<le> c * n ^ d + e"
  shows "f \<in> O_of (Comp_Poly d)"
proof -
  show ?thesis unfolding O_of_def
    by (rule, rule bexI[of _ "\<lambda> n. c * n ^ d + e"], insert ge, auto)
qed

lemma O_of_sum: assumes f1: "f1 \<in> O_of cc"
  and f2: "f2 \<in> O_of cc"
  shows "(\<lambda> n. f1 n + f2 n) \<in> O_of cc"
proof (cases cc)
  case (Comp_Poly d)
  from f1[unfolded Comp_Poly O_of_def, simplified]
  obtain c1 e1 where b1: "\<And> n. f1 n \<le> c1 * n ^ d + e1" by auto
  from f2[unfolded Comp_Poly O_of_def, simplified]
  obtain c2 e2 where b2: "\<And> n. f2 n \<le> c2 * n ^ d + e2" by auto
  show ?thesis unfolding Comp_Poly
  proof (rule O_of_polyI)
    fix n
    have "f1 n + f2 n \<le> c1 * n ^ d + e1 + c2 * n ^ d + e2"
      using b1[of n] b2[of n] by auto
    also have "... = (c1 + c2) * n ^ d + (e1 + e2)" by (simp add: field_simps)
    finally show "f1 n + f2 n \<le> (c1 + c2) * n ^ d + (e1 + e2)" .
  qed
qed

fun degree :: "complexity_class \<Rightarrow> nat"
  where "degree (Comp_Poly d) = d"

instantiation complexity_class :: ord
begin
fun less_eq_complexity_class where "less_eq_complexity_class (x::complexity_class) (y :: complexity_class) = (degree x \<le> degree y)"
fun less_complexity_class where "less_complexity_class (x::complexity_class) (y :: complexity_class) = (degree x < degree y)"
instance ..
end

instantiation complexity_class :: "show" begin
  definition "shows_prec d (c::complexity_class) \<equiv> (if degree c = 0 then shows ''O(1)'' else if degree c = 1 then shows ''O(n)'' else shows ''O(n^'' +@+ shows (degree c) +@+ shows '')'')"
  definition "shows_list (bs::complexity_class list) \<equiv> shows_list_aux shows bs"

  lemma assoc_complexity_class: "shows_prec d (b::complexity_class) r @ s = shows_prec d b (r @ s)"
    by (unfold shows_prec_complexity_class_def shows_string_def,simp)
  lemma assoc_complexity_class_elt: "\<forall>(b::complexity_class)\<in>set bs. shows_prec d b r @ s = shows_prec d b (r @ s)"
    using assoc_complexity_class by auto
  instance by 
    (intro_classes,rule assoc_complexity_class,unfold shows_list_complexity_class_def)
    (rule shows_list_aux_assoc[OF assoc_complexity_class_elt])
end


lemma O_of_mono: assumes c: "cc \<le> cc'"
  shows "O_of cc \<subseteq> O_of cc'"
proof -
  obtain d where cc: "cc = Comp_Poly d" by (cases cc, auto)
  obtain d' where cc': "cc' = Comp_Poly d'" by (cases cc', auto)
  from c cc cc' have "d \<le> d'" by simp
  from O_of_poly_mono_deg[OF this]
  show ?thesis unfolding cc cc' .
qed

definition deriv_bound_rel :: "'a ars \<Rightarrow> (nat \<Rightarrow> 'a set) \<Rightarrow> (nat \<Rightarrow> nat) \<Rightarrow> bool"
  where "deriv_bound_rel r as f \<equiv> \<forall> n a. a \<in> as n \<longrightarrow> deriv_bound r a (f n)"


lemma deriv_bound_rel_empty[simp]: "deriv_bound_rel {} as f"
  unfolding deriv_bound_rel_def by simp

definition deriv_bound_rel_class :: "'a ars \<Rightarrow> (nat \<Rightarrow> 'a set) \<Rightarrow> complexity_class \<Rightarrow> bool"
  where "deriv_bound_rel_class r as cpx \<equiv> \<exists> f. f \<in> O_of cpx \<and> deriv_bound_rel r as f"

lemma deriv_bound_rel_class_polyI[intro]: 
  assumes bound: "\<And> n a. a \<in> as n \<longrightarrow> deriv_bound r a (c * n ^ d + e)"
  shows "deriv_bound_rel_class r as (Comp_Poly d)"
  unfolding deriv_bound_rel_class_def
  by (rule exI[of _ "\<lambda> n. c * n ^ d + e"], rule conjI[OF O_of_polyI[OF le_refl]],
  insert bound, unfold deriv_bound_rel_def, auto)


lemma deriv_bound_rel_class_empty[simp]: "deriv_bound_rel_class {} as cpx"
proof (cases cpx)
  case (Comp_Poly d)
  show ?thesis
    unfolding deriv_bound_rel_class_def Comp_Poly
    by (rule exI, rule conjI, rule O_of_polyI[OF le_refl], simp)
qed

lemma deriv_bound_rel_step_class_union: assumes s1: "deriv_bound_rel_class (rel_step (s1,w \<union> s2)) as cc"
  and s2: "deriv_bound_rel_class (rel_step (s2,w \<union> s1)) as cc"
  shows "deriv_bound_rel_class (rel_step (s1 \<union> s2, w)) as cc"
proof -
  let ?r1 = "rel_step (s1,w \<union> s2)"
  let ?r2 = "rel_step (s2,w \<union> s1)"
  let ?r  = "rel_step (s1 \<union> s2, w)"
  from s1[unfolded deriv_bound_rel_class_def]
  obtain f1 where f1: "f1 \<in> O_of cc" and b1: "deriv_bound_rel ?r1 as f1" by auto
  from s2[unfolded deriv_bound_rel_class_def]
  obtain f2 where f2: "f2 \<in> O_of cc" and b2: "deriv_bound_rel ?r2 as f2" by auto
  obtain f where f: "f = (\<lambda> n. f1 n + f2 n)" by auto
  have fcc: "f \<in> O_of cc" using O_of_sum[OF f1 f2] unfolding f .
  show ?thesis unfolding deriv_bound_rel_class_def
  proof(intro exI conjI, rule fcc)
    show "deriv_bound_rel ?r as f"
      unfolding deriv_bound_rel_def
    proof (intro allI impI)
      fix n a
      assume a: "a \<in> as n"
      from b1[unfolded deriv_bound_rel_def] a
      have b1: "deriv_bound ?r1 a (f1 n)" by simp
      from b2[unfolded deriv_bound_rel_def] a
      have b2: "deriv_bound ?r2 a (f2 n)" by simp
      show "deriv_bound ?r a (f n)" 
        unfolding deriv_bound_def
      proof
        assume "\<exists> b. (a,b) \<in> ?r ^^ Suc (f n)"
        then obtain b where ab: "(a,b) \<in> ?r ^^ Suc (f n)" by auto
        from ab[unfolded rel_pow_fun_conv] obtain as where
          as0: "as 0 = a"
          and steps: "\<And> i. i < Suc (f n) \<Longrightarrow> (as  i, as (Suc i)) \<in> ?r" by auto
        {
          fix i
          assume "i < Suc (f n)"
          from steps[OF this, unfolded rel_step.simps]
          obtain b c where asb: "(as i, b) \<in> w^*"  and bc: "(b,c) \<in> s1 \<union> s2" and cas: "(c,as (Suc i)) \<in> w^*" by auto
          let ?p = "(as i, as (Suc i))"
          {
            fix s
            have "(as i, b) \<in> (w \<union> s)^*"
              by (rule set_mp[OF _ asb], regexp)
          } note asb = this
          {
            fix s
            have "(c, as (Suc i)) \<in> (w \<union> s)^*"
              by (rule set_mp[OF _ cas], regexp)
          } note cas = this
          from bc
          have "?p \<in> ?r1 \<and> ?p \<in> (w \<union> s1)^* \<or> ?p \<in> ?r2 \<and> ?p \<in> (w \<union> s2)^*"
          proof
            assume bc: "(b,c) \<in> s1"
            have s: "?p \<in> ?r1" unfolding rel_step.simps using asb[of s2] bc cas[of s2] by auto
            have ns: "?p \<in> (w \<union> s1)^* O (w \<union> s1) O (w \<union> s1)^* " using asb[of s1] bc cas[of s1] by auto
            have "?p \<in> (w \<union> s1)^*" by (rule set_mp[OF _ ns], regexp)
            with s show ?thesis by blast
          next
            assume bc: "(b,c) \<in> s2"
            have s: "?p \<in> ?r2" unfolding rel_step.simps using asb[of s1] bc cas[of s1] by auto
            have ns: "?p \<in> (w \<union> s2)^* O (w \<union> s2) O (w \<union> s2)^* " using asb[of s2] bc cas[of s2] by auto
            have "?p \<in> (w \<union> s2)^*" by (rule set_mp[OF _ ns], regexp)
            with s show ?thesis by blast
          qed
        } note steps = this
        {
          fix i
          assume i: "i \<le> Suc (f n)"
          let ?p = "\<lambda> i i1 i2 b1 b2. (a,b1) \<in> ?r1 ^^ i1 \<and> (b1,as i) \<in> (w \<union> s2)^* \<and> (a,b2) \<in> ?r2 ^^ i2 \<and> (b2,as i) \<in> (w \<union> s1)^* \<and> i1 + i2 = i"
          from i have "\<exists> i1 i2 b1 b2. ?p i i1 i2 b1 b2"
          proof (induct i)
            case 0
            show ?case
            proof (intro exI)
              show "?p 0 0 0 a a" unfolding as0 by simp
            qed
          next
            case (Suc i)
            from Suc(2) have "i \<le> Suc (f n)" by simp
            from Suc(1)[OF this] obtain i1 i2 b1 b2 where p: "?p i i1 i2 b1 b2" by blast
            hence ab1: "(a,b1) \<in> ?r1 ^^ i1"
              and b1ai: "(b1,as i) \<in> (w \<union> s2)^*"
              and ab2: "(a,b2) \<in> ?r2 ^^ i2"
              and b2ai: "(b2,as i) \<in> (w \<union> s1)^*" 
              and i: "i = i1 + i2" by auto
            let ?a = "(as i, as (Suc i))"
            from Suc(2) have "i < Suc (f n)" by simp
            from steps[OF this] show ?case
            proof
              assume "?a \<in> ?r1 \<and> ?a \<in> (w \<union> s1)^*"
              hence s: "?a \<in> ?r1" and ns: "?a \<in> (w \<union> s1)^*" by auto
              from b1ai s have s: "(b1, as (Suc i)) \<in> (w \<union> s2)^* O ?r1" by auto
              have "(b1, as (Suc i)) \<in> ?r1" 
                by (rule set_mp[OF _ s], unfold rel_step.simps, regexp)
              from rel_pow_Suc_I[OF ab1 this]
              have s: "(a, as (Suc i)) \<in> ?r1 ^^ Suc i1" .
              from b2ai ns have ns: "(b2, as (Suc i)) \<in> (w \<union> s1)^*" by auto
              have "?p (Suc i) (Suc i1) i2 (as (Suc i)) b2" using s ns i ab2 by auto
              thus ?thesis by blast
            next
              assume one: "?a \<in> ?r2 \<and> ?a \<in> (w \<union> s2)^*"
              hence s: "?a \<in> ?r2" and ns: "?a \<in> (w \<union> s2)^*" by auto
              from b2ai s have s: "(b2, as (Suc i)) \<in> (w \<union> s1)^* O ?r2" by auto
              have "(b2, as (Suc i)) \<in> ?r2" 
                by (rule set_mp[OF _ s], unfold rel_step.simps, regexp)
              from rel_pow_Suc_I[OF ab2 this]
              have s: "(a, as (Suc i)) \<in> ?r2 ^^ Suc i2" .
              from b1ai ns have ns: "(b1, as (Suc i)) \<in> (w \<union> s2)^*" by auto
              have "?p (Suc i) i1 (Suc i2) b1 (as (Suc i))" using s ns i ab1 by auto
              thus ?thesis by blast
            qed
          qed
          hence "\<exists> i1 i2 b1 b2. (a,b1) \<in> ?r1 ^^ i1 \<and> (a,b2) \<in> ?r2 ^^ i2 \<and> i1 + i2 = i" by blast
        }
        from this[OF le_refl]
        obtain i1 i2 b1 b2 where steps1: "(a,b1) \<in> ?r1 ^^ i1" and steps2: "(a,b2) \<in> ?r2 ^^ i2" and i: "i1 + i2 = Suc (f n)" by blast
        from deriv_bound_steps[OF steps1 b1] deriv_bound_steps[OF steps2 b2] have "i1 + i2 \<le> f1 n + f2 n" by simp
        with i[unfolded f] show False by simp
      qed
    qed
  qed
qed            


datatype ('f,'v) complexity_measure = 
  Derivational_Complexity "('f \<times> nat) list" | (* signature *)
  Runtime_Complexity "('f \<times> nat) list" "('f \<times> nat) list" (* constructors, defined *) 

fun terms_of_nat :: "('f,'v)complexity_measure \<Rightarrow> nat \<Rightarrow> ('f,'v)term set"
  where "terms_of_nat (Derivational_Complexity F) n = { t. funas_term t \<subseteq> set F \<and> size t \<le> n}"
      | "terms_of_nat (Runtime_Complexity C D) n = { t. is_Fun t \<and> the (root t) \<in> set D \<and> (\<forall> s \<in> set (args t). funas_term s \<subseteq> set C) \<and> size t \<le> n}"

definition deriv_bound_measure_class :: "('f,'v)term ars \<Rightarrow> ('f,'v)complexity_measure \<Rightarrow> complexity_class \<Rightarrow> bool"
  where "deriv_bound_measure_class r cm cpx \<equiv> deriv_bound_rel_class r (terms_of_nat cm) cpx"

lemma deriv_bound_measure_class_empty[simp]: "deriv_bound_measure_class {} cm cc"
  unfolding deriv_bound_measure_class_def by simp

lemma deriv_bound_rel_step_measure_class_union: assumes s1: "deriv_bound_measure_class (rel_step (s1,w \<union> s2)) as cc"
  and s2: "deriv_bound_measure_class (rel_step (s2,w \<union> s1)) as cc"
  shows "deriv_bound_measure_class (rel_step (s1 \<union> s2, w)) as cc"
  using s1 s2
  unfolding deriv_bound_measure_class_def
  by (rule deriv_bound_rel_step_class_union)

lemma deriv_bound_rel_mono: assumes r: "r \<subseteq> r'" and as: "\<And> n. as n \<subseteq> as' n"
  and f: "\<And> n. f' n \<le> f n"
  and d: "deriv_bound_rel r' as' f'"
  shows "deriv_bound_rel r as f" unfolding deriv_bound_rel_def
proof (intro allI impI)
  fix n a 
  assume "a \<in> as n"
  with as have "a \<in> as' n" by auto
  from d[unfolded deriv_bound_rel_def, rule_format, OF this]
  have "deriv_bound r' a (f' n)" .
  from deriv_bound_mono[OF this f]
  have "deriv_bound r' a (f n)" .
  from deriv_bound_subset[OF r this]
  show "deriv_bound r a (f n)" .
qed

lemma deriv_bound_rel_class_mono: assumes r: "r \<subseteq> r'" and as: "\<And> n. as n \<subseteq> as' n"
  and cc: "O_of cc' \<subseteq> O_of cc"
  and d: "deriv_bound_rel_class r' as' cc'"
  shows "deriv_bound_rel_class r as cc" unfolding deriv_bound_rel_class_def
proof -
  from d[unfolded deriv_bound_rel_class_def] obtain f where f: "f \<in> O_of cc'"
    and d: "deriv_bound_rel r' as' f" by auto
  from f cc have f: "f \<in> O_of cc" by auto
  from deriv_bound_rel_mono[OF r, of as as', OF as le_refl d] f
  show "\<exists> f. f \<in> O_of cc \<and> deriv_bound_rel r as f" by auto
qed

lemma deriv_bound_measure_class_mono: assumes r: "r \<subseteq> r'" and cm: "\<And> n. terms_of_nat cm n \<subseteq> terms_of_nat cm' n"
  and cc: "O_of cc' \<subseteq> O_of cc"
  and d: "deriv_bound_measure_class r' cm' cc'"
  shows "deriv_bound_measure_class r cm cc" unfolding deriv_bound_measure_class_def
  by (rule deriv_bound_rel_class_mono[OF r cm cc], rule d[unfolded deriv_bound_measure_class_def])

lemma runtime_subset_derivational: "terms_of_nat (Runtime_Complexity C D) n \<subseteq> terms_of_nat (Derivational_Complexity (C @ D)) n"
proof (rule)
  fix t
  assume t: "t \<in> terms_of_nat (Runtime_Complexity C D) n"
  from t have n: "size t \<le> n" by simp
  from t obtain f ts where f: "t = Fun f ts" by (cases t, auto)
  from t f have D: "(f,length ts) \<in> set D" and C: "\<And> t. t \<in> set ts \<Longrightarrow> funas_term t \<subseteq> set C" by auto
  from D C have "funas_term t \<subseteq> set (C @ D)" unfolding f by auto
  with t show "t \<in> terms_of_nat (Derivational_Complexity (C @ D)) n" by simp
qed  

end


